/*
* SACD Decoder plugin
* Copyright (c) 2011 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "../ATLHelpers/ATLHelpers.h"
#include "resource.h"

#include <stdint.h>
#include <foobar2000.h>
#include "asio_output.h"
#include "sacd_reader.h"
#include "sacd_disc.h"
#include "sacd_dsdiff.h"
#include "sacd_setup.h"
#include "dst_decoder_foo.h"
#include "dsdpcm_converter.h"
#include "dsd2pcm_pack.h"

using namespace pfc;

enum media_type_t {UNK_TYPE = 0, ISO_TYPE = 1, DSDIFF_TYPE = 2};

static const uint32_t UPDATE_STATS_MS = 500;
static const int BITRATE_AVGS = 16;

void console_fprintf(FILE* file, char* fmt, ...) {
	va_list vl;
	va_start(vl, fmt);
	console::printfv(fmt, vl);
	va_end(vl);
}

int get_sacd_channel_map_from_loudspeaker_config(int loudspeaker_config) {
	int sacd_channel_map;
	switch (loudspeaker_config) {
	case 0:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right;
		break;
	case 3:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_left | audio_chunk::channel_back_right;
		break;
	case 4:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_lfe | audio_chunk::channel_back_left | audio_chunk::channel_back_right;
		break;
	default:
		sacd_channel_map = 0;
		break;
	}
	return sacd_channel_map;
}

int get_sacd_channel_map_from_channels(int channels) {
	int sacd_channel_map;
	switch (channels) {
	case 2:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right;
		break;
	case 5:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_left | audio_chunk::channel_back_right;
		break;
	case 6:
		sacd_channel_map = audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_lfe | audio_chunk::channel_back_left | audio_chunk::channel_back_right;
		break;
	default:
		sacd_channel_map = 0;
		break;
	}
	return sacd_channel_map;
}

CoASIOProxy        g_proxy;
asio_output_t      g_asio_output;
dsdpcm_converter_t g_dsdpcm_playback;

class playback_handler_t : public play_callback_static {
	bool   m_seek_made;
	double m_seek_time;
public:
	playback_handler_t() {}
	void on_playback_starting(play_control::t_track_command p_command, bool p_paused) {}
	void on_playback_new_track(metadb_handle_ptr p_track) {
		g_asio_output.playback_start();
	}
	void on_playback_stop(play_control::t_stop_reason p_reason) {
		g_asio_output.playback_stop();
		g_proxy.request_mode(ASIO_MODE_PCM);
	}
	void on_playback_seek(double p_time) {
		g_asio_output.playback_seek();
	}
	void on_playback_pause(bool p_state) {
		g_asio_output.playback_pause(p_state);
	}
	void on_playback_edited(metadb_handle_ptr p_track) {}
	void on_playback_dynamic_info(const file_info& p_info) {}
	void on_playback_dynamic_info_track(const file_info& p_info) {}
	void on_playback_time(double p_time) {}
	void on_volume_change(float p_new_val) {}
	unsigned get_flags() {
		return flag_on_playback_new_track | flag_on_playback_stop | flag_on_playback_seek | flag_on_playback_pause;
	}
};

static play_callback_static_factory_t<playback_handler_t> g_playback_sacd_factory;

class input_sacd_t {
	media_type_t          media_type;
	service_ptr_t<file>   media_file;
	sacd_reader_t*        sacd_reader;
	unsigned              flags;
	int                   sacd_bitrate[BITRATE_AVGS];
	int                   sacd_bitrate_idx;
	int                   sacd_bitrate_sum;
	array_t<uint8_t>      dsd_buf;
	int                   dsd_buf_size;
	array_t<uint8_t>      dst_buf;
	int                   dst_buf_size;
	array_t<audio_sample> pcm_buf;
	dst_decoder_t*        dst_decoder;
	dsdpcm_converter_t    dsdpcm_convert;
	dsd2pcm_packager_t    dsd2pcm_packager;
	uint32_t              info_update_time_ms;
	int                   pcm_out_channels;
	unsigned int          pcm_out_channel_map;
	int                   pcm_out_samplerate;
	int                   pcm_out_bits_per_sample;
	int                   pcm_out_samples;
	bool                  dsd_use_asio;
	int                   dsd_samplerate;
public:
	input_sacd_t() {
		sacd_reader = NULL;
		dst_decoder_create_mt(&dst_decoder, DST_DECODER_THREADS);
		info_update_time_ms = 0;
		g_proxy.request_mode(ASIO_MODE_PCM);
	}

	virtual ~input_sacd_t() {
		if (sacd_reader) {
			delete sacd_reader;
		}
		dst_decoder_destroy_mt(dst_decoder);
		g_proxy.request_mode(ASIO_MODE_PCM);
	}

	void open(service_ptr_t<file> p_filehint, const char* p_path, t_input_open_reason p_reason, abort_callback& p_abort) {
		string_extension ext(p_path);
		media_type = UNK_TYPE;
		if (stricmp_utf8(ext, "ISO") == 0) {
			media_type = ISO_TYPE;
		}
		if (stricmp_utf8(ext, "DAT") == 0) {
			media_type = ISO_TYPE;
		}
		if (stricmp_utf8(ext, "DFF") == 0) {
			media_type = DSDIFF_TYPE;
		}
		if (media_type == UNK_TYPE) {
			throw exception_io_unsupported_format();
		}
		media_file = p_filehint; 
		input_open_file_helper(media_file, p_path, p_reason, p_abort);
		sacd_reader = NULL;
		switch (media_type) {
		case ISO_TYPE:
			sacd_reader = new sacd_disc_t;
			if (!sacd_reader) {
				throw exception_overflow();
			}
			break;
		case DSDIFF_TYPE:
			sacd_reader = new sacd_dsdiff_t;
			if (!sacd_reader) {
				throw exception_overflow();
			}
			break;
		default:
			throw exception_io_data();
			break;
		}
		if (!sacd_reader->open(media_file)) {
			throw exception_io_data();
		}
		pcm_out_samplerate = CSACDPreferences::get_samplerate();
		pcm_out_bits_per_sample = 24;
	}

	t_uint32 get_subsong_count() {
		switch (media_type) {
		case ISO_TYPE:
			return reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_TWOCH) + reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_MULCH);
			break;
		case DSDIFF_TYPE:
			return 1;
			break;
		}
		return 0;
	}

	t_uint32 get_subsong(t_uint32 p_index) {
		return p_index;
	}

	void get_info(t_uint32 p_subsong, file_info& p_info, abort_callback& p_abort) {
		if (media_type == ISO_TYPE) {
			scarletbook_area_t* area = 0;
			uint8_t twoch_count = reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_TWOCH);
			if (p_subsong < twoch_count) {
				area = reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_area(AREA_TWOCH);
			}
			else {
				p_subsong -= twoch_count;
				if (p_subsong < reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_MULCH)) {
					area = reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_area(AREA_MULCH);
				}
			}
			if (!area) {
				return;
			}
			area_tracklist_time_duration_t duration = area->area_tracklist_time->duration[p_subsong];
			p_info.set_length(duration.minutes * 60.0 + duration.seconds * 1.0 + duration.frames / 75.0);
			p_info.info_set_int("samplerate", DSDxFs64);
			p_info.info_set_int("channels", area->area_toc->channel_count);
			p_info.info_set_int("bitspersample", pcm_out_bits_per_sample);
			string_formatter(codec);
			codec << (area->area_toc->frame_format == FRAME_FORMAT_DST ? "DST" : "DSD");
			codec << DSDxFs64 / 44100;
			p_info.info_set("codec", codec);
			p_info.info_set("encoding", "lossless");
			p_info.info_set_bitrate(((t_int64)(DSDxFs64 * area->area_toc->channel_count) + 500) / 1000);
		}
		if (media_type == DSDIFF_TYPE) {
			p_info.set_length(sacd_reader->get_duration());
			p_info.info_set_int("samplerate", sacd_reader->get_samplerate());
			p_info.info_set_int("channels", sacd_reader->get_channels());
			p_info.info_set_int("bitspersample", pcm_out_bits_per_sample);
			string_formatter(codec);
			codec << (sacd_reader->is_dst() ? "DST" : "DSD");
			codec << sacd_reader->get_samplerate() / 44100;
			p_info.info_set("codec", codec);
			p_info.info_set("encoding", "lossless");
			p_info.info_set_bitrate(((t_int64)(sacd_reader->get_samplerate() * sacd_reader->get_channels()) + 500) / 1000);
		}
		sacd_reader->get_info(p_subsong, p_info);
	}

	t_filestats get_file_stats(abort_callback& p_abort) {
		return media_file->get_stats(p_abort);
	}

	void decode_initialize(t_uint32 p_subsong, unsigned p_flags, abort_callback& p_abort) {
		flags = p_flags;
		if (media_type == ISO_TYPE) {
			uint8_t twoch_count = reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_TWOCH);
			if (p_subsong < twoch_count) {
				if (!reinterpret_cast<sacd_disc_t*>(sacd_reader)->set_track(AREA_TWOCH, (uint8_t)p_subsong, 0)) {
					throw exception_io();
				}
			}
			else {
				p_subsong -= twoch_count;
				if (p_subsong < reinterpret_cast<sacd_disc_t*>(sacd_reader)->get_track_count(AREA_MULCH)) {
					if (!reinterpret_cast<sacd_disc_t*>(sacd_reader)->set_track(AREA_MULCH, (uint8_t)p_subsong, 0)) {
						throw exception_io();
					}
				}
			}
		}
		if (dst_decoder_init_mt(dst_decoder, sacd_reader->get_channels()) != 0) {
			throw exception_io();
		}
		dsd_samplerate = sacd_reader->get_samplerate();
		pcm_out_channels = sacd_reader->get_channels();
		dst_buf_size = dsd_buf_size = dsd_samplerate / 8 / 75 * pcm_out_channels; 
		dsd_buf.set_size(dst_decoder->thread_count * dsd_buf_size);
		dst_buf.set_size(dst_decoder->thread_count * dst_buf_size);
		pcm_out_channel_map = get_sacd_channel_map_from_loudspeaker_config(sacd_reader->get_loudspeaker_config());
		pcm_out_channel_map = get_sacd_channel_map_from_channels(pcm_out_channels);
		pcm_out_samples = pcm_out_samplerate / 75;
		pcm_buf.set_size(pcm_out_channels * pcm_out_samples);
		memset(sacd_bitrate, 0, sizeof(sacd_bitrate));
		sacd_bitrate_idx = 0;
		sacd_bitrate_sum = 0;
		dsd_use_asio = false;
		if (flags & input_flag_playback) {
			g_dsdpcm_playback.init(pcm_out_channels, dsd_samplerate, pcm_out_samplerate, (conv_type_t)CSACDPreferences::get_converter_mode());
			g_dsdpcm_playback.set_gain((float)CSACDPreferences::get_volume());
			dsd_use_asio = CSACDPreferences::in_dsd_mode();
			if (dsd_use_asio) {
				if (!g_asio_output.create(pcm_out_channels, dsd_samplerate)) {
					throw exception_io_not_found();
				}
				g_proxy.request_mode(ASIO_MODE_DSD);
			}
			else {
				g_proxy.request_mode(ASIO_MODE_PCM);
			}
			g_proxy.request_dsd_samplerate(dsd_samplerate);
			g_proxy.request_pcm_samplerate(pcm_out_samplerate);
		}
		else {
			dsdpcm_convert.init(pcm_out_channels, dsd_samplerate, pcm_out_samplerate, (conv_type_t)CSACDPreferences::get_converter_mode());
			dsdpcm_convert.set_gain((float)CSACDPreferences::get_volume());
		}
	}

	bool decode_run(audio_chunk& p_chunk, abort_callback& p_abort) {
		uint8_t* dsd_data;
		uint8_t* dst_data;
		size_t dsd_size = 0;
		size_t dst_size = 0;
		for (;;) {
			dsd_data = dsd_buf.get_ptr() + dsd_buf_size * dst_decoder->slot_nr;
			dst_data = dst_buf.get_ptr() + dst_buf_size * dst_decoder->slot_nr;
			if (sacd_reader->read_frame(dst_data, &dst_size)) {
				if (dst_size > 0) {
					if (sacd_reader->is_dst()) {
						dst_decoder_decode_mt(dst_decoder, dst_data, dst_size, &dsd_data, &dsd_size);
					}
					else {
						dsd_data = dst_data;
						dsd_size = dst_size;
					}
					sacd_bitrate_idx = (++sacd_bitrate_idx) % BITRATE_AVGS;
					sacd_bitrate_sum -= sacd_bitrate[sacd_bitrate_idx];
					sacd_bitrate[sacd_bitrate_idx] = dst_size * 8 * 75;
					sacd_bitrate_sum += sacd_bitrate[sacd_bitrate_idx];
					if (dsd_size > 0) {
						if (flags & input_flag_playback) {
							g_dsdpcm_playback.convert(dsd_data, pcm_buf.get_ptr(), dsd_size);
						}
						else {
							dsdpcm_convert.convert(dsd_data, pcm_buf.get_ptr(), dsd_size);
						}
						if (dsd_use_asio) {
							g_asio_output.set_data(dsd_data, dsd_size);
						}
						p_chunk.set_data(pcm_buf.get_ptr(), pcm_out_samples, pcm_out_channels, pcm_out_samplerate, pcm_out_channel_map);
						return true;
					}
				}
			}
			else {
				break;
			}
		}
		dsd_data = NULL;
		dst_data = NULL;
		dst_size = 0;
		dst_decoder_decode_mt(dst_decoder, dst_data, dst_size, &dsd_data, &dsd_size);
		if (dsd_size > 0) {
			if (flags & input_flag_playback) {
				g_dsdpcm_playback.convert(dsd_data, pcm_buf.get_ptr(), dsd_size);
			}
			else {
				dsdpcm_convert.convert(dsd_data, pcm_buf.get_ptr(), dsd_size);
			}
			if (dsd_use_asio) {
				g_asio_output.set_data(dsd_data, dsd_size);
			}
			p_chunk.set_data(pcm_buf.get_ptr(), pcm_out_samples, pcm_out_channels, pcm_out_samplerate, pcm_out_channel_map);
			return true;
		}
		return false;
	}

	void decode_seek(double p_seconds, abort_callback& p_abort) {
		uint64_t offset = (uint64_t)(sacd_reader->get_size() * p_seconds / sacd_reader->get_duration());
		if (!sacd_reader->seek(offset)) {
			throw exception_io();
		}
	}

	bool decode_can_seek() {
		return media_file->can_seek();
	}

	bool decode_get_dynamic_info(file_info& p_info, double& p_timestamp_delta) {
		DWORD curr_time_ms = GetTickCount();
		if (info_update_time_ms + (DWORD)UPDATE_STATS_MS >= curr_time_ms) {
			return false;
		}
		info_update_time_ms = curr_time_ms;
		p_info.info_set_bitrate_vbr(((t_int64)(sacd_bitrate_sum / BITRATE_AVGS) + 500) / 1000);
		string_formatter(codec);
		codec << (sacd_reader->is_dst() ? "DST" : "DSD");
		codec << sacd_reader->get_samplerate() / 44100;
		p_info.info_set("codec", codec);
		if (dsd_use_asio && g_proxy.get_mode() == ASIO_MODE_DSD) {
			p_info.info_set_int("samplerate", sacd_reader->get_samplerate());
		}
		else {
			p_info.info_set_int("samplerate", pcm_out_samplerate);
		}
		return true;
	}
	
	bool decode_get_dynamic_info_track(file_info& p_info, double& p_timestamp_delta) {
		return false;
	}
	
	void decode_on_idle(abort_callback & p_abort) {
		media_file->on_idle(p_abort);
	}

	void retag_set_info(t_uint32 p_subsong, const file_info& p_info, abort_callback& p_abort) {
		throw exception_not_implemented();
	}
	
	void retag_commit(abort_callback& p_abort) {
	}

	static bool g_is_our_content_type(const char* p_content_type) {
		return false;
	}
	
	static bool g_is_our_path(const char* p_path, const char* p_ext) {
		return (stricmp_utf8(p_ext, "ISO") == 0 || stricmp_utf8(p_ext, "DAT") == 0) && sacd_disc_t::g_is_sacd(p_path) || stricmp_utf8(p_ext, "DFF") == 0;
	}
};

static input_factory_t<input_sacd_t> g_input_sacd_factory;

class initquit_sacd_t : public initquit {
public:
	virtual void on_init() {
		g_asio_output.hook(true);
	}
	
	virtual void on_quit() {
		g_asio_output.hook(false);
	}
};

static initquit_factory_t<initquit_sacd_t> g_initquit_sacd_factory;

DECLARE_COMPONENT_VERSION("Super Audio CD Decoder", "0.3.4" , "Super Audio CD Decoder Input PlugIn.\n\nCopyright (c) 2011 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>");
DECLARE_FILE_TYPE("SACD files", "*.DAT;*.DFF;*.ISO");
