/*
* SACD Decoder plugin
* Copyright (c) 2011 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <math.h>
#include <memory.h>
#include "dsdpcm_converter_real.h"

void dsdpcm_fir_r::init(ctable_r* fir_ctables, int fir_length, int channels, int decimation) {
	this->fir_ctables = fir_ctables;
	this->fir_length  = CTABLES(fir_length);
	this->channels    = channels;
	this->decimation  = decimation / 8;
	free();
	int buf_length = 2 * this->fir_length * sizeof(uint8_t);
	for (int ch = 0; ch < this->channels; ch++) {
		this->fir_buffer[ch] = (uint8_t*)_aligned_malloc(buf_length, SSE_ALIGN);
		memset(this->fir_buffer[ch], 0x55, buf_length);
	}
	fir_index = 0;
}

void dsdpcm_fir_r::free() {
	for (int ch = 0; ch < DSDPCM_MAX_CHANNELS; ch++) {
		if (fir_buffer[ch]) {
			_aligned_free(fir_buffer[ch]);
			fir_buffer[ch] = NULL;
		}
	}
}

int dsdpcm_fir_r::run(uint8_t* dsd_data, real_t* pcm_data, int dsd_samples) {
	int pcm_samples = dsd_samples / decimation;
	for (int sample = 0; sample < pcm_samples; sample += channels) {
		for (int ch = 0; ch < channels; ch++) {
			pcm_data[sample + ch] = 0;
			for (int j = 0; j < fir_length; j++) {
				pcm_data[sample + ch] += fir_ctables[j][fir_buffer[ch][fir_index + j]];
			}
		}
		for (int i = 0; i < decimation; i++) {
			for (int ch = 0; ch < channels; ch++) {
				fir_buffer[ch][fir_index + fir_length] = fir_buffer[ch][fir_index] = *(dsd_data++);
			}
			fir_index = (++fir_index) % fir_length;
		}
	}
	return pcm_samples;
}

void pcmpcm_fir_r::init(real_t* fir_coefs, int fir_length, int channels, int decimation) {
	this->fir_coefs  = fir_coefs;
	this->fir_length = SSE_ASIZE(fir_length, sizeof(real_t));
	this->channels   = channels;
	this->decimation = decimation;
	free();
	int buf_length = 2 * this->fir_length * sizeof(real_t);
	for (int ch = 0; ch < this->channels; ch++) {
		this->fir_buffer[ch] = (real_t*)_aligned_malloc(buf_length, SSE_ALIGN);
		memset(this->fir_buffer[ch], 0, buf_length);
	}
	fir_index = 0;
}

void pcmpcm_fir_r::free() {
	for (int ch = 0; ch < DSDPCM_MAX_CHANNELS; ch++) {
		if (fir_buffer[ch]) {
			_aligned_free(fir_buffer[ch]);
			fir_buffer[ch] = NULL;
		}
	}
}

int pcmpcm_fir_r::run(real_t* pcm_data, real_t* out_data, int pcm_samples) {
	int out_samples = pcm_samples / decimation;
	for (int sample = 0; sample < out_samples; sample += channels) {
		for (int ch = 0; ch < channels; ch++) {
			out_data[sample + ch] = (real_t)0;
			for (int j = 0; j < fir_length; j++) {
				out_data[sample + ch] += fir_coefs[j] * fir_buffer[ch][fir_index + j];
			}
		}
		for (int i = 0; i < decimation; i++) {
			for (int ch = 0; ch < channels; ch++) {
				fir_buffer[ch][fir_index + fir_length] = fir_buffer[ch][fir_index] = *(pcm_data++);
			}
			fir_index = (++fir_index) % fir_length;
		}
	}
	return out_samples;
}

int pcmpcm_fir_r::run_sse(real_t* pcm_data, real_t* out_data, int pcm_samples) {
	int out_samples = pcm_samples / decimation;
	for (int sample = 0; sample < out_samples; sample += channels) {
		for (int ch = 0; ch < channels; ch++) {
			__m128 s = _mm_setzero_ps();
			for (int j = 0; j < fir_length; j += SSE_ALIGN / sizeof(real_t)) {
				s = _mm_add_ps(s, _mm_mul_ps(_mm_load_ps(&fir_coefs[j]), _mm_loadu_ps(&fir_buffer[ch][fir_index + j])));
			}
			s = _mm_add_ps(s, _mm_shuffle_ps(s, s, _MM_SHUFFLE(1, 0, 3, 2)));
			s = _mm_add_ps(s, _mm_shuffle_ps(s, s, _MM_SHUFFLE(0, 1, 2, 3)));
			_mm_store_ss(&out_data[sample + ch], s);
		}
		for (int i = 0; i < decimation; i++) {
			for (int ch = 0; ch < channels; ch++) {
				fir_buffer[ch][fir_index + fir_length] = fir_buffer[ch][fir_index] = *(pcm_data++);
			}
			fir_index = (++fir_index) % fir_length;
		}
	}
	return out_samples;
}

real_t dsdpcm_converter_r::dsd_fir1_8_ctables[CTABLES(DSDFIR1_8_LENGTH)][256];
real_t dsdpcm_converter_r::dsd_fir1_16_ctables[CTABLES(DSDFIR1_16_LENGTH)][256];
real_t dsdpcm_converter_r::pcm_fir2_2_coefs[SSE_ASIZE(PCMFIR2_2_LENGTH, sizeof(real_t))];
real_t dsdpcm_converter_r::pcm_fir3_2_coefs[SSE_ASIZE(PCMFIR3_2_LENGTH, sizeof(real_t))];

dsdpcm_converter_r::dsdpcm_converter_r() {
	conv_type = DSDPCM_CONV_REAL;
}

dsdpcm_converter_r::~dsdpcm_converter_r() {
}

int dsdpcm_converter_r::init(int channels, int dsd_samplerate, int pcm_samplerate) {
	static bool preinitialized = false;
	if (!preinitialized) {
		preinit();
		preinitialized = true;
	}
	int CPUInfo[4];
  __cpuid(CPUInfo, 1);
	sse2_enabled = (CPUInfo[3] & (1L << 26)) ? true : false;
	this->channels = channels;
	this->dsd_samplerate = dsd_samplerate;
	this->pcm_samplerate = pcm_samplerate;
	switch (dsd_samplerate) {
	case DSDxFs64:
		switch (pcm_samplerate) {
		case DSDxFs1:
			conv_mode = DSD64_44100;
			break;
		case DSDxFs2:
			conv_mode = DSD64_88200;
			break;
		case DSDxFs4:
			conv_mode = DSD64_176400;
			break;
		case DSDxFs8:
			conv_mode = DSD64_352800;
			break;
		default:
			return -2;
		}
		break;
	case DSDxFs128:
		switch (pcm_samplerate) {
		case DSDxFs1:
			conv_mode = DSD128_44100;
			break;
		case DSDxFs2:
			conv_mode = DSD128_88200;
			break;
		case DSDxFs4:
			conv_mode = DSD128_176400;
			break;
		case DSDxFs8:
			conv_mode = DSD128_352800;
			break;
		default:
			return -2;
		}
		break;
	default:
		return -1;
		break;
	}
	float dB_gain_adjust = 18.0f;
	switch (conv_mode) {
	case DSD64_44100:
		dsd_fir1.init(dsd_fir1_16_ctables, DSDFIR1_16_LENGTH, channels, 16);
		pcm_fir2a.init(pcm_fir2_2_coefs, PCMFIR2_2_LENGTH, channels, 2);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD64_88200:
		dsd_fir1.init(dsd_fir1_8_ctables, DSDFIR1_8_LENGTH, channels, 8);
		pcm_fir2a.init(pcm_fir2_2_coefs, PCMFIR2_2_LENGTH, channels, 2);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD64_176400:
		dsd_fir1.init(dsd_fir1_8_ctables, DSDFIR1_8_LENGTH, channels, 8);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD64_352800:
		dsd_fir1.init(dsd_fir1_8_ctables, DSDFIR1_8_LENGTH, channels, 8);
		dB_gain_adjust -= 1.0f;
		break;
	case DSD128_44100:
		dsd_fir1.init(dsd_fir1_16_ctables, DSDFIR1_16_LENGTH, channels, 16);
		pcm_fir2a.init(pcm_fir2_2_coefs, PCMFIR2_2_LENGTH, channels, 2);
		pcm_fir2b.init(pcm_fir2_2_coefs, PCMFIR2_2_LENGTH, channels, 2);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD128_88200:
		dsd_fir1.init(dsd_fir1_16_ctables, DSDFIR1_16_LENGTH, channels, 16);
		pcm_fir2a.init(pcm_fir2_2_coefs, PCMFIR2_2_LENGTH, channels, 2);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD128_176400:
		dsd_fir1.init(dsd_fir1_16_ctables, DSDFIR1_16_LENGTH, channels, 16);
		pcm_fir3.init(pcm_fir3_2_coefs, PCMFIR3_2_LENGTH, channels, 2);
		break;
	case DSD128_352800:
		dsd_fir1.init(dsd_fir1_16_ctables, DSDFIR1_16_LENGTH, channels, 16);
		break;
	}
	gain0 = pow(10.0f, dB_gain_adjust / 20.0f);
	gain = (real_t)gain0;
	return 0;
}

int dsdpcm_converter_r::convert(uint8_t* dsd_data, int32_t* pcm_data, int dsd_samples) {
	int pcm_samples;
	if (sse2_enabled) {
		pcm_samples = convert_internal_sse(dsd_data, pcm_tempo, dsd_samples);
	}
	else {
		pcm_samples = convert_internal(dsd_data, pcm_tempo, dsd_samples);
	}
	for (int i = 0; i < pcm_samples; i++) {
		pcm_data[i] = (int32_t)pcm_tempo[i];
	}
	return pcm_samples;
}

int dsdpcm_converter_r::convert(uint8_t* dsd_data, float* pcm_data, int dsd_samples) {
	int pcm_samples;
	if (sse2_enabled) {
		pcm_samples = convert_internal_sse(dsd_data, pcm_tempo, dsd_samples);
	}
	else {
		pcm_samples = convert_internal(dsd_data, pcm_tempo, dsd_samples);
	}
	for (int i = 0; i < pcm_samples; i++) {
		pcm_data[i] = (float)(gain * pcm_tempo[i]);
	}
	return pcm_samples;
}

void dsdpcm_converter_r::set_gain(float dB_gain) {
	gain = (real_t)(gain0 * pow(10.0f, dB_gain / 20.0f));
}

int dsdpcm_converter_r::set_ctables(int32_t* fir_coefs, int fir_length, ctable_r* fir_ctables) {
	int ctables = CTABLES(fir_length);
	for (int ct = 0; ct < ctables; ct++) {
		int k = fir_length - ct * 8;
		if (k > 8) {
			k = 8;
		}
		if (k < 0) {
			k = 0;
		}
		for (int i = 0; i < 256; i++) {
			real_t cvalue = (real_t)0;
			for (int j = 0; j < k; j++) {
				cvalue += (real_t)(((i >> j) & 1) * 2 - 1) * (real_t)fir_coefs[fir_length - 1 - (ct * 8 + 7 - j)] * NORM;
			}
			fir_ctables[ct][i] = cvalue;
		}
	}
	return ctables;
}

void dsdpcm_converter_r::preinit() {
	set_ctables((int32_t*)DSDFIR1_8_COEFS, DSDFIR1_8_LENGTH, (ctable_r*)dsd_fir1_8_ctables);
	set_ctables((int32_t*)DSDFIR1_16_COEFS, DSDFIR1_16_LENGTH, (ctable_r*)dsd_fir1_16_ctables);
	for (int i = 0; i < PCMFIR2_2_LENGTH; i++) {
		pcm_fir2_2_coefs[i] = (real_t)PCMFIR2_2_COEFS[PCMFIR2_2_LENGTH - 1 - i] * NORM;
	}
	for (int i = PCMFIR2_2_LENGTH; i < SSE_ASIZE(PCMFIR2_2_LENGTH, sizeof(real_t)); i++) {
		pcm_fir2_2_coefs[i] = (real_t)0;
	}
	for (int i = 0; i < PCMFIR3_2_LENGTH; i++) {
		pcm_fir3_2_coefs[i] = (real_t)PCMFIR3_2_COEFS[PCMFIR3_2_LENGTH - 1 - i] * NORM;
	}
	for (int i = PCMFIR3_2_LENGTH; i < SSE_ASIZE(PCMFIR3_2_LENGTH, sizeof(real_t)); i++) {
		pcm_fir3_2_coefs[i] = (real_t)0;
	}
}

int dsdpcm_converter_r::convert_internal(uint8_t* dsd_data, real_t* pcm_data, int dsd_samples) {
	int pcm_samples;
	switch (conv_mode) {
	case DSD64_44100:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD64_88200:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD64_176400:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir3.run(pcm_temp1, pcm_data, pcm_samples);
		break;
	case DSD64_352800:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_data, dsd_samples);
		break;
	case DSD128_44100:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir2b.run(pcm_temp2, pcm_temp3, pcm_samples);
		pcm_samples = pcm_fir3.run(pcm_temp3, pcm_data, pcm_samples);
		break;
	case DSD128_88200:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD128_176400:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir3.run(pcm_temp1, pcm_data, pcm_samples);
		break;
	case DSD128_352800:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_data, dsd_samples);
		break;
	}
	return pcm_samples;
}

int dsdpcm_converter_r::convert_internal_sse(uint8_t* dsd_data, real_t* pcm_data, int dsd_samples) {
	int pcm_samples;
	switch (conv_mode) {
	case DSD64_44100:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run_sse(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD64_88200:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run_sse(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD64_176400:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp1, pcm_data, pcm_samples);
		break;
	case DSD64_352800:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_data, dsd_samples);
		break;
	case DSD128_44100:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run_sse(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir2b.run_sse(pcm_temp2, pcm_temp3, pcm_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp3, pcm_data, pcm_samples);
		break;
	case DSD128_88200:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir2a.run_sse(pcm_temp1, pcm_temp2, pcm_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp2, pcm_data, pcm_samples);
		break;
	case DSD128_176400:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_temp1, dsd_samples);
		pcm_samples = pcm_fir3.run_sse(pcm_temp1, pcm_data, pcm_samples);
		break;
	case DSD128_352800:
		pcm_samples = dsd_fir1.run(dsd_data, pcm_data, dsd_samples);
		break;
	}
	return pcm_samples;
}
